"""
Helper functions for downloading sample model output data.
"""
from pathlib import Path

from parfive import Downloader

download_dir = Path(__file__).parent / '..' / '..' / 'data'


def mas_helio():
    """
    Get some MAS heliospheric data files. These are taken from CR2210, which
    is used for PSP data comparisons in the documentation examples.
    """
    mas_helio_dir = download_dir / 'mas_helio'
    mas_helio_dir.mkdir(parents=True, exist_ok=True)
    base_url = 'http://www.predsci.com/data/runs/cr2210-medium/hmi_masp_mas_std_0201/helio/{var}002.hdf'

    # Create a downloader to queue the files to be downloaded
    dl = Downloader()

    vars = ['rho', 'vr', 'br']
    for var in vars:
        file = mas_helio_dir / f'{var}002.hdf'
        if file.exists():
            continue
        else:
            remote_file = base_url.format(var=var)
            dl.enqueue_file(remote_file, path=mas_helio_dir)

    # Download the files
    if dl.queued_downloads > 0:
        dl.download()
    return mas_helio_dir.resolve()



"""
Sampling data from a 3D model
=============================

In this example we'll see how to sample a 3D model output at arbitrary points
within the model domain.
"""
###############################################################################
# First, load the required modules.
from psipy.model import MASOutput
from psipy.data import sample_data

import astropy.constants as const
import astropy.units as u

import matplotlib.pyplot as plt
import numpy as np

###############################################################################
# Load a set of MAS output files, and get the number density variable from the
# model run.
mas_path = sample_data.mas_helio()
model = MASOutput(mas_path)
rho = model['rho']

###############################################################################
# Choose a set of 1D points to interpolate the model output at.
#
# Here we keep a constant radius, and a set of longitudes that go all the way
# from 0 to 360 degrees. Then we choose two different, but close latitude
# values, and plot the results.
#
# As expected, the values at 0 and 360 degrees are the same, and the profiles
# are similar, but different, due to the small difference in latitude.
fig, ax = plt.subplots()

npoints = 1000
r = 50 * np.ones(npoints) * const.R_sun
lon = np.linspace(0, 360, npoints) * u.deg

for latitude in [0, 1] * u.deg:
    lat = latitude * np.ones(npoints)
    samples = rho.sample_at_coords(lon, lat, r)

    ax.plot(lon, samples, label='lat = ' + str(latitude))

ax.legend()
ax.set_xlim(0, 360)
ax.set_ylim(bottom=0)
ax.set_xlabel('Longitude (deg)')
ax.set_ylabel(r'$\rho$ (cm$^{-3}$)')
ax.set_xticks([0, 90, 180, 270, 360])
plt.show()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import sphinx_rtd_theme

# -- Project information -----------------------------------------------------

project = 'psipy'
copyright = '2020, Predictive Science Inc.'
author = 'David Stansby'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx_automodapi.automodapi',
              'sphinx_gallery.gen_gallery',
              "sphinx_rtd_theme",
              'numpydoc',
              'sphinx.ext.intersphinx']

numpydoc_show_class_members = False
sphinx_gallery_conf = {
    'examples_dirs': '../examples',   # path to your example scripts
    'gallery_dirs': 'auto_examples',  # path to where to save gallery generated output
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Raise an error for missing doc references
nitpicky = True


intersphinx_mapping = {'python': ('https://docs.python.org/3', None),
                       'xarray': ('http://xarray.pydata.org/en/stable', None),
                       'astropy': ('https://docs.astropy.org/en/stable/', None),
                       'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
                       'heliopy': ('https://docs.heliopy.org/en/stable/', None)
                       }

default_role = 'any'
automodapi_inheritance_diagram = False
automodsumm_inherited_members = True



"""
This file includes global test configuration.

In particular, it defines the location where test data is available.
"""
from pathlib import Path

import pytest

from psipy.data import sample_data

test_data_dir = (Path(__file__) / '..' / '..' / 'data').resolve()


@pytest.fixture(scope="module", params=['mas_helio', 'mas_hdf5'])
def mas_directory(request):
    if request.param == 'mas_helio':
        # Check for and download data if not present
        directory = sample_data.mas_helio()
    else:
        # Directories with MAS outputs
        directory = test_data_dir / request.param
    if not directory.exists():
        pytest.xfail(f'Could not find MAS data directory at {directory}')

    return directory


@pytest.fixture(scope="module")
def pluto_directory():
    # Directories with PLUTO outputs
    directory = test_data_dir / 'pluto'
    if not directory.exists():
        pytest.xfail(f'Could not find PLUTO data directory at {directory}')

    return directory



import astropy.constants as const
import astropy.units as u

from .base import ModelOutput
from psipy.io import read_pluto_files, get_pluto_variables

__all__ = ['PLUTOOutput']


class PLUTOOutput(ModelOutput):
    """
    The results from a single run of PLUTO.

    This is a storage object that contains a number of `Variable` objects. It
    is designed to be used like::

        pluto_output = PLUTOOutput('directory')
        br = pluto_output['br']

    Notes
    -----
    Variables are loaded on demand. To see the list of available variables
    use `PLUTOOutput.variables`, and to see the list of already loaded variables
    use `PLUTOOutput.loaded_variables`.
    """
    def get_unit(self, var):
        return u.dimensionless_unscaled, 1

    def get_variables(self):
        return get_pluto_variables(self.path)

    def load_file(self, var):
        return read_pluto_files(self.path, var)



import astropy.constants as const
import astropy.units as u

from .base import ModelOutput
from psipy.io import read_mas_file, get_mas_variables

__all__ = ['MASOutput']


# A mapping from unit names to their units, and factors the data needs to be
# multiplied to get them into these units.
_vunit = [u.km / u.s, 481.37]
_bunit = [u.G, 2.205]
_junit = [u.A / u.m**2, 2.267e4]
_mas_units = {'vr': _vunit,
              'vt': _vunit,
              'vp': _vunit,
              'va': _vunit,
              'br': _bunit,
              'bt': _bunit,
              'bp': _bunit,
              'bmag': _bunit,
              'rho': [u.N / u.cm**3, 1.67e-16 / 1.67e-24],
              't': [u.K, 2.804e7],
              'p': [u.Pa, 3.875717e-2],
              'jr': _junit,
              'jt': _junit,
              'jp': _junit
              }


class MASOutput(ModelOutput):
    """
    The results from a single run of MAS.

    This is a storage object that contains a number of `Variable` objects. It
    is designed to be used like::

        mas_output = MASOutput('directory')
        br = mas_output['br']

    Notes
    -----
    Variables are loaded on demand. To see the list of available variables
    use `MASOutput.variables`, and to see the list of already loaded variables
    use `MASOutput.loaded_variables`.
    """
    def get_unit(self, var):
        return _mas_units[var]

    def get_variables(self):
        return get_mas_variables(self.path)

    def load_file(self, var):
        return read_mas_file(self.path, var)

    def __repr__(self):
        return f'psipy.model.mas.MASOutput("{self.path}")'

    def __str__(self):
        return f"MAS output in directory {self.path}"



import copy
import textwrap

import astropy.constants as const
import astropy.units as u
import numpy as np
from scipy import interpolate
import xarray as xr

import psipy.visualization as viz


__all__ = ['Variable']


class Variable:
    """
    A single scalar variable.

    This class primarily contains methods for plotting data. It can be created
    with any `xarray.DataArray` that has ``['theta', 'phi', 'r']`` fields.

    Parameters
    ----------
    data : xarray.DataArray
        Variable data.
    name : str
        Variable name.
    unit : astropy.units.Quantity
        Variable unit.
    """
    def __init__(self, data, name, unit):
        self._data = data
        # Sort the data once now for any interpolation later
        self._data = self._data.sortby(['phi', 'theta', 'r'])
        self.name = name
        self._unit = unit

    def __str__(self):
        return textwrap.dedent(f'''
        Variable
        --------
        Name: {self.name}
        Grid size: {self._data.shape}
        ''')

    @property
    def data(self):
        """
        DataFrame with the data.
        """
        return self._data

    @property
    def unit(self):
        """
        Units of the scalar data.
        """
        return self._unit

    @property
    def r_coords(self):
        """
        Radial coordinate values.
        """
        return self._data.coords['r'].values

    @property
    def theta_coords(self):
        """
        Latitude coordinate values.
        """
        return self._data.coords['theta'].values

    @property
    def phi_coords(self):
        """
        Longitude coordinate values.
        """
        return self._data.coords['phi'].values

    def radial_normalized(self, radial_exponent):
        r"""
        Return a radially normalised copy of this variable.

        Multiplies the variable by :math:`(r / r_{\odot})^{\gamma}`,
        where :math:`\gamma` = ``radial_exponent`` is the given exponent.

        Parameters
        ----------
        radial_exponent : float

        Returns
        -------
        Variable
        """
        r = self.data.coords['r']
        rsun_au = float(u.AU / const.R_sun)
        data = self.data * (r * rsun_au)**radial_exponent
        units = self.unit * (u.AU)**radial_exponent
        name = self.name + f' $r^{radial_exponent}$'
        return Variable(data, name, units)

    # Methods for radial cuts
    @staticmethod
    def _setup_radial_ax(ax):
        if ax is None:
            ax = plt.gca()
        return ax

    @staticmethod
    def _format_radial_ax(ax):
        # Plot formatting
        ax.set_aspect('equal')
        ax.set_xlim(0, 2 * np.pi)
        ax.set_ylim(-np.pi / 2, np.pi / 2)
        viz.clear_axes_labels(ax)

        # Tick label formatting
        viz.set_theta_formatters(ax)
        ax.set_xticks(np.deg2rad(np.linspace(0, 360, 7, endpoint=True)))
        ax.set_yticks(np.deg2rad(np.linspace(-90, 90, 7, endpoint=True)))

    def plot_radial_cut(self, i, ax=None, **kwargs):
        """
        Plot a radial cut.

        Parameters
        ----------
        i : int
            Index at which to slice the data.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.pcolormesh`.
        """
        ax = self._setup_radial_ax(ax)

        kwargs = self._set_cbar_label(kwargs, self.unit.to_string('latex'))
        # Take slice of data, and plot
        sliced = self.data.isel(r=i)
        sliced.plot(x='phi', y='theta', ax=ax, **kwargs)

        # Plot formatting
        r = sliced['r'].values
        ax.set_title(f'{self.name}, r={r:.2f}' + r'$R_{\odot}$')
        self._format_radial_ax(ax)

    def contour_radial_cut(self, i, levels, ax=None, **kwargs):
        """
        Plot contours on a radial cut.

        Parameters
        ----------
        i : int
            Index at which to slice the data.
        levels : list
            List of levels to contour.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.contour`.
        """
        ax = self._setup_radial_ax(ax)
        sliced = self.data.isel(r=i)
        # Need to save a copy of the title to reset it later, since xarray
        # tries to set it's own title that we don't want
        title = ax.get_title()
        xr.plot.contour(sliced, x='phi', y='theta', ax=ax,
                        levels=levels, **kwargs)
        ax.set_title(title)
        self._format_radial_ax(ax)

    def plot_phi_cut(self, i, ax=None, **kwargs):
        """
        Plot a phi cut.

        Parameters
        ----------
        i : int
            Index at which to slice the data.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.pcolormesh`.
        """
        ax = viz.setup_polar_ax(ax)

        kwargs = self._set_cbar_label(kwargs, self.unit.to_string('latex'))
        # Take slice of data and plot
        sliced = self.data.isel(phi=i)
        sliced.plot(x='theta', y='r', ax=ax, **kwargs)
        viz.format_polar_ax(ax)
        phi = np.rad2deg(sliced['phi'].values)
        ax.set_title(f'{self.name}, ' + r'$\phi$= ' + f'{phi:.2f}' + '$^{\circ}$')

    def contour_phi_cut(self, i, levels, ax=None, **kwargs):
        """
        Plot contours on a phi cut.

        Parameters
        ----------
        i : int
            Index at which to slice the data.
        levels : list
            List of levels to contour.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.contour`.
        """
        ax = viz.setup_polar_ax(ax)
        sliced = self.data.isel(phi=i)
        # Need to save a copy of the title to reset it later, since xarray
        # tries to set it's own title that we don't want
        title = ax.get_title()
        xr.plot.contour(sliced, x='theta', y='r', ax=ax,
                        levels=levels, **kwargs)
        viz.format_polar_ax(ax)
        ax.set_title(title)

    @property
    def _equator_theta_idx(self):
        """
        The theta index of the solar equator.
        """
        return (self.data.shape[1] - 1) // 2

    # Methods for equatorial cuts
    def plot_equatorial_cut(self, ax=None, **kwargs):
        """
        Plot an equatorial cut.

        Parameters
        ----------
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.pcolormesh`.
        """
        ax = viz.setup_polar_ax(ax)
        kwargs = self._set_cbar_label(kwargs, self.unit.to_string('latex'))
        # Get data slice
        sliced = self.data.isel(theta=self._equator_theta_idx)
        # Plot
        sliced.plot(x='phi', y='r', ax=ax, **kwargs)
        # Plot formatting
        viz.format_equatorial_ax(ax)
        theta = np.rad2deg(sliced['theta'].values)
        ax.set_title(f'{self.name}, equatorial plane')

    def contour_equatorial_cut(self, levels, ax=None, **kwargs):
        """
        Plot contours on an equatorial cut.

        Parameters
        ----------
        levels : list
            List of levels to contour.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.contour`.
        """
        ax = viz.setup_polar_ax(ax)
        sliced = self.data.isel(theta=self._equator_theta_idx)
        # Need to save a copy of the title to reset it later, since xarray
        # tries to set it's own title that we don't want
        title = ax.get_title()
        xarray.plot.contour(sliced, x='phi', y='r', ax=ax,
                            levels=levels, **kwargs)
        viz.format_equatorial_ax(ax)
        ax.set_title(title)

    @staticmethod
    def _set_cbar_label(kwargs, label):
        """
        Set the colobar label with units.
        """
        # Copy kwargs to prevent modifying them inplace
        kwargs = copy.deepcopy(kwargs)
        # Set the colobar label with units
        cbar_kwargs = kwargs.pop('cbar_kwargs', {})
        cbar_kwargs['label'] = cbar_kwargs.pop('label', label)
        kwargs['cbar_kwargs'] = cbar_kwargs
        return kwargs

    @u.quantity_input
    def sample_at_coords(self, lon, lat, r):
        """
        Sample this variable along a 1D trajectory of coordinates.

        Parameters
        ----------
        lon : astropy.units.Quantity
            Longitudes.
        lat : astropy.units.Quantity
            Latitudes.
        r : astropy.units.Quantity
            Radial distances.

        Returns
        -------
        astropy.units.Quantity
            The sampled data.

        Notes
        -----
        Linear interpolation is used to interpoalte between cells. See the
        docstring of `scipy.interpolate.interpn` for more information.
        """
        points = [self.data.coords[dim].values for dim in ['phi', 'theta', 'r']]
        values = self.data.values
        # Check that coordinates are increasing
        if not np.all(np.diff(points[0]) >= 0):
            raise RuntimeError('Longitude coordinates are not monotonically increasing')
        if not np.all(np.diff(points[1]) >= 0):
            raise RuntimeError('Latitude coordinates are not monotonically increasing')
        if not np.all(np.diff(points[2]) > 0):
            raise RuntimeError('Radial coordinates are not monotonically increasing')

        # Pad phi points so it's possible to interpolate all the way from
        # 0 to 360 deg
        points[0] = np.append(points[0], points[0][0] + 2 * np.pi)
        values = np.append(values, values[0:1, ...], axis=0)

        xi = np.column_stack([lon.to_value(u.rad),
                              lat.to_value(u.rad),
                              r.to_value(const.R_sun)])

        values_x = interpolate.interpn(points, values, xi)
        return values_x * self._unit



import os

import numpy as np
import pyhdf.SD as h4
import h5py as h5


__all__ = ['read_hdf4', 'read_hdf5']


class HDF4File:
    """
    A context manager for automatically opening/closing HDF4 files
    """
    def __init__(self, file_name):
        file_name = str(file_name)
        if not os.path.exists(file_name):
            raise FileNotFoundError(f'Could not find {file_name}')
        self.file_obj = h4.SD(file_name)

    def __enter__(self):
        return self.file_obj

    def __exit__(self, type, value, traceback):
        self.file_obj.end()


def read_hdf4(path, sds_id='Data-Set-2'):
    """
    Read a HDF4 file.

    Reads a single dataset from a single HDF4 file, returning the scalar data
    and associated coordinates.

    Parameters
    ----------
    path :
        Path to the file.
    sds_id : str, optional
        ID of the dataset to get.

    Returns
    -------
    data : ndarray
        Scalar data.
    coords : list of ndarray
        Cooordinates values along each axis of the data.
    """
    # Load the HDF4 file
    # In all PSI files the data is stored in "Data-Set-2"
    with HDF4File(path) as sd_id:
        sds_id = sd_id.select('Data-Set-2')

        # Get the scalar data
        data = sds_id.get()
        # Get coordinate information
        coords = [sds_id.dim(i).getscale() for i in range(np.ndim(data))]

    return data, coords


def read_hdf5(path, dataset_name='Data'):
    """
    Read a HDF5 file.

    Reads a single dataset from a single HDF5 file, returning the scalar data
    and associated coordinates.

    Parameters
    ----------
    path :
        Path to the file.
    dataset_name : str, optional
        ID of the dataset to get.

    Returns
    -------
    data : ndarray
        Scalar data.
    coords : list of ndarray
        Cooordinates values along each axis of the data.
    """
    with h5.File(path, 'r') as hdf5_file:
        # Get the scalar data
        data = np.array(hdf5_file[dataset_name])
        # Get coordinate information
        coords = [np.array(hdf5_file[dataset_name].dims[i][0]) for i in
                  range(np.ndim(data))]
        coords = coords[::-1]

    return data, coords



"""
Plotting equatorial slices
==========================

This example shows how to plot slices in the equatorial plane from a MAS model
output.
"""
###############################################################################
# First, load the required modules.
from psipy.model import MASOutput
from psipy.data import sample_data

import matplotlib.pyplot as plt

###############################################################################
# Next, load a set of MAS output files. You will need to change this line to
# point to a folder with MAS files in them.
mas_path = sample_data.mas_helio()
model = MASOutput(mas_path)

###############################################################################
# Each MAS model contains a number of variables. The variable names can be
# accessed using the ``.variables`` attribute.
print(model.variables)

###############################################################################
# Set parameters for plotting. This line will give us a horizontal
# errorbar underneath the plots.
cbar_kwargs = {'orientation': 'horizontal'}

###############################################################################
# Plot the slices
ax = plt.subplot(projection='polar')
model['vr'].plot_equatorial_cut(ax=ax, cbar_kwargs=cbar_kwargs)

plt.show()
